from .base import Query
from .methods import (
    Alter,
    Avg,
    Count,
    Create,
    CreateIndex,
    Delete,
    DropIndex,
    Exists,
    Insert,
    Max,
    Min,
    Objects,
    Raw,
    Select,
    Sum,
    TableExists,
    Update,
)
